/*
 * Copyright (C) 2016 - 2018  (See AUTHORS)
 *
 * This file is part of Owl.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package owl.automaton;

import static org.hamcrest.Matchers.containsInAnyOrder;
import static org.hamcrest.Matchers.empty;
import static org.hamcrest.Matchers.equalTo;
import static org.hamcrest.Matchers.is;
import static org.junit.Assert.assertThat;
import static owl.automaton.DefaultImplementations.getReachableStates;

import com.google.common.collect.Iterators;
import com.google.common.collect.Lists;
import java.util.BitSet;
import java.util.List;
import java.util.Objects;
import org.junit.Test;
import owl.automaton.Automaton.Property;
import owl.automaton.acceptance.AllAcceptance;
import owl.automaton.acceptance.BuchiAcceptance;
import owl.automaton.edge.Edge;
import owl.automaton.edge.LabelledEdge;
import owl.factories.ValuationSetFactory;
import owl.run.DefaultEnvironment;

public class HashMapAutomatonTest {
  private static final ValuationSetFactory FACTORY;

  static {
    FACTORY = DefaultEnvironment.annotated().factorySupplier().getValuationSetFactory(List.of("a"));
  }

  @Test
  public void testEmptyAutomaton() {
    HashMapAutomaton<?, ?> automaton = new HashMapAutomaton<>(FACTORY, AllAcceptance.INSTANCE);
    assertThat(automaton.is(Property.COMPLETE), is(false));
    assertThat(automaton.is(Property.DETERMINISTIC), is(true));
    assertThat(AutomatonUtil.getIncompleteStates(automaton).keySet(), empty());
    assertThat(automaton.initialStates(), empty());
    assertThat(automaton.states(), empty());
  }

  @SuppressWarnings("unchecked")
  @Test
  public void testRemapAcceptanceDuplicate() {
    // Test that edges which are equal after a remapAcceptance call get merged.

    MutableAutomaton<String, ?> automaton = new HashMapAutomaton<>(FACTORY, AllAcceptance.INSTANCE);

    automaton.addInitialState("1");
    automaton.addInitialState("2");

    BitSet other = new BitSet();
    other.set(0);
    automaton.addEdge("1", new BitSet(), Edge.of("1", 1));
    automaton.addEdge("1", other, Edge.of("2"));
    automaton.addEdge("1", new BitSet(), Edge.of("2", 1));
    automaton.updateEdges((state, edge) -> {
      if (edge.successor().equals("2")) {
        return Edge.of(edge.successor());
      }

      if (edge.successor().equals("1")) {
        return null;
      }

      return edge;
    });
    automaton.trim();

    assertThat(automaton.labelledEdges("1"), containsInAnyOrder(
      LabelledEdge.of(Edge.of("2"), FACTORY.universe())));
  }

  @SuppressWarnings("unchecked")
  @Test
  public void testSimpleAutomaton() {
    // Test various parts of the implementation on a simple, two-state automaton
    MutableAutomaton<String, ?> automaton = new HashMapAutomaton<>(FACTORY,
      BuchiAcceptance.INSTANCE);

    automaton.addState("1");
    automaton.addState("2");
    automaton.trim();

    assertThat(automaton.states(), empty());
    assertThat(automaton.states(), equalTo(getReachableStates(automaton)));

    automaton.addInitialState("1");

    assertThat(automaton.onlyInitialState(), is("1"));
    assertThat(automaton.initialStates(), containsInAnyOrder("1"));
    assertThat(automaton.states(), containsInAnyOrder("1"));
    assertThat(automaton.states(), equalTo(getReachableStates(automaton)));

    // Add edge
    automaton.addEdge("1", FACTORY.universe(), Edge.of("2"));

    assertThat(automaton.edge("1", new BitSet()), is(Edge.of("2")));
    assertThat(automaton.edges("1", new BitSet()), containsInAnyOrder(Edge.of("2")));
    assertThat(automaton.labelledEdges("1"), containsInAnyOrder(
      LabelledEdge.of("2", FACTORY.universe())));
    assertThat(automaton.edge("1", new BitSet()), is(Edge.of("2")));
    assertThat(automaton.states(), containsInAnyOrder("1", "2"));
    assertThat(automaton.states(), equalTo(getReachableStates(automaton)));

    assertThat(automaton.is(Property.DETERMINISTIC), is(true));
    assertThat(automaton.is(Property.SEMI_DETERMINISTIC), is(true));
    assertThat(automaton.is(Property.LIMIT_DETERMINISTIC), is(true));

    // Add duplicate edge
    automaton.addEdge("1", FACTORY.of(new BitSet()), Edge.of("2"));

    assertThat(automaton.labelledEdges("1").size(), is(1));
    assertThat(automaton.edge("1", new BitSet()), is(Edge.of("2")));
    assertThat(automaton.states(), containsInAnyOrder("1", "2"));
    assertThat(automaton.states(), equalTo(getReachableStates(automaton)));

    assertThat(automaton.is(Property.DETERMINISTIC), is(true));
    assertThat(automaton.is(Property.SEMI_DETERMINISTIC), is(true));
    assertThat(automaton.is(Property.LIMIT_DETERMINISTIC), is(true));

    // Add edge with different acceptance
    automaton.addEdge("1", FACTORY.of(new BitSet()), Edge.of("2", 0));

    assertThat(automaton.edges("1", new BitSet()),
      containsInAnyOrder(Edge.of("2"), Edge.of("2", 0)));
    assertThat(Iterators.size(automaton.labelledEdges("1").iterator()), is(2));
    assertThat(automaton.states(), containsInAnyOrder("1", "2"));
    assertThat(automaton.states(), equalTo(getReachableStates(automaton)));

    assertThat(automaton.is(Property.DETERMINISTIC), is(false));
    assertThat(automaton.is(Property.SEMI_DETERMINISTIC), is(false));
    assertThat(automaton.is(Property.LIMIT_DETERMINISTIC), is(true));

    automaton.addEdge("1", FACTORY.universe(), Edge.of("1", 0));

    assertThat(automaton.edges("1", new BitSet()),
      containsInAnyOrder(Edge.of("1", 0), Edge.of("2"), Edge.of("2", 0)));
    assertThat(automaton.states(), containsInAnyOrder("1", "2"));
    assertThat(automaton.states(), equalTo(getReachableStates(automaton)));

    assertThat(automaton.is(Property.DETERMINISTIC), is(false));
    assertThat(automaton.is(Property.SEMI_DETERMINISTIC), is(false));
    assertThat(automaton.is(Property.LIMIT_DETERMINISTIC), is(false));

    automaton.addInitialState("2");

    assertThat(automaton.states(), containsInAnyOrder("1", "2"));
    assertThat(automaton.states(), equalTo(getReachableStates(automaton)));
    assertThat(AutomatonUtil.getIncompleteStates(automaton).keySet(), containsInAnyOrder("2"));

    assertThat(automaton.is(Property.DETERMINISTIC), is(false));
    assertThat(automaton.is(Property.SEMI_DETERMINISTIC), is(false));
    assertThat(automaton.is(Property.LIMIT_DETERMINISTIC), is(false));

    automaton.updateEdges(automaton.states(), (state, edge) -> edge.withAcceptance(key -> 2));
    automaton.trim();

    assertThat(automaton.edges("1", new BitSet()),
      containsInAnyOrder(Edge.of("1", 2), Edge.of("2"), Edge.of("2", 2)));
    assertThat(automaton.edges("2", new BitSet()), empty());

    automaton.updateEdges((state, edge) -> Objects.equals(state, "1")
      ? Edge.of(edge.successor(), 0)
      : Edge.of(edge.successor(), 1));
    automaton.trim();

    for (LabelledEdge<String> successorEdge : automaton.labelledEdges("1")) {
      assertThat(Lists.newArrayList(successorEdge.edge.acceptanceSetIterator()),
        containsInAnyOrder(0));
    }

    for (LabelledEdge<String> successorEdge : automaton.labelledEdges("2")) {
      assertThat(Lists.newArrayList(successorEdge.edge.acceptanceSetIterator()),
        containsInAnyOrder(1));
    }

    assertThat("Automaton is inconsistent.",
      ((HashMapAutomaton<?, ?>) automaton).checkConsistency(), is(true));
  }
}
