#include "Automaton.h"

#include <iostream>

namespace aut {

Automaton::Automaton(owl::Automaton _automaton) :
    automaton(std::move(_automaton)),
    node_type(initNodeType()),
    max_color(initMaxColor()),
    default_color(initDefaultColor()),
    parity_type(initParityType()),
    has_lock(false),
    complete(false)
{
    // reserve some space in the vector to avoid initial resizing, which needs locking for the parallel construction
    successors.reserve(4096);
}

color_t Automaton::initMaxColor() const {
    switch (automaton.acceptance()) {
        case owl::PARITY_MIN_EVEN:
        case owl::PARITY_MIN_ODD:
            return automaton.acceptance_set_count() - 1;
        default:
            return 1;
    }
}

color_t Automaton::initDefaultColor() const {
    switch (automaton.acceptance()) {
        case owl::CO_SAFETY:
            return 1;
        case owl::SAFETY:
            return 0;
        default:
            return 0;
    }
}

NodeType Automaton::initNodeType() const {
    switch (automaton.acceptance()) {
        case owl::PARITY_MIN_EVEN:
        case owl::PARITY_MIN_ODD:
            return NodeType::PARITY;
        case owl::BUCHI:
            return NodeType::BUCHI;
        case owl::CO_BUCHI:
            return NodeType::CO_BUCHI;
        case owl::SAFETY:
            return NodeType::WEAK;
        case owl::CO_SAFETY:
            return NodeType::WEAK;
        default:
            throw std::invalid_argument("unsupported acceptance");
    }
}

Parity Automaton::initParityType() const {
    switch (automaton.acceptance()) {
        case owl::PARITY_MIN_EVEN:
        case owl::BUCHI:
        case owl::SAFETY:
        case owl::CO_SAFETY:
            return Parity::EVEN;
        case owl::PARITY_MIN_ODD:
        case owl::CO_BUCHI:
            return Parity::ODD;
        default:
            throw std::invalid_argument("unsupported acceptance");
    }
}

color_t Automaton::getMaxColor() const {
    return max_color;
}

NodeType Automaton::getNodeType() const {
    return node_type;
}

Parity Automaton::getParityType() const {
    return parity_type;
}

void Automaton::add_new_query(node_id_t query) {
    std::unique_lock<std::mutex> lock(query_mutex);
    queries.push(query);
    query_mutex.unlock();
    change.notify_all();
}

void Automaton::mark_complete() {
    complete = true;
    change.notify_all();
}

void Automaton::wait_for_query(node_id_t& query) {
    std::unique_lock<std::mutex> lock(query_mutex);
    while(!complete && queries.empty()) {
        change.wait(lock);
    }
    if (!queries.empty()) {
        query = queries.front();
        queries.pop();
    }
}

void Automaton::add_new_states() {
    while (!complete) {
        node_id_t query = NODE_BOTTOM;
        wait_for_query(query);
        if (query != NODE_BOTTOM) {
            add_successors(query);
        }
    }
}

void Automaton::add_successors(node_id_t local_state) {
    if (local_state >= successors.size()) {
        /*
        if (successors.capacity() >= local_state + 1) {
            successors_mutex.lock();
            has_lock = true;
        }
        */
        successors.resize(local_state + 1);
        /*
        if (has_lock) {
            successors_mutex.unlock();
            has_lock = false;
        }
        */
    }

    if (successors[local_state].empty()) {
        if (node_type == NodeType::WEAK) {
            const std::vector<int> local_successors = automaton.successors(local_state);
            successors[local_state].reserve(local_successors.size());
            for (const auto successor : local_successors) {
                node_id_t new_state;
                if (successor >= 0) {
                    new_state = successor;
                }
                else if (successor == -1) {
                    new_state = NODE_BOTTOM;
                }
                else if (successor == -2) {
                    new_state = NODE_TOP;
                }
                else {
                    throw std::invalid_argument("invalid successor for automaton");
                }
                successors[local_state].push_back(new_state);
            }
        }
        else {
            const std::vector<owl::Edge> edges = automaton.edges(local_state);
            successors[local_state].reserve(2*edges.size());
            for (const auto& edge : edges) {
                node_id_t new_state;
                color_t color;
                if (edge.successor >= 0) {
                    new_state = edge.successor;
                    if (node_type == NodeType::BUCHI || node_type == NodeType::CO_BUCHI) {
                        if (edge.colour == -1) {
                            color = 1;
                        }
                        else {
                            color = 0;
                        }
                    }
                    else if (edge.colour == -1) {
                        color = max_color;
                    }
                    else {
                        color = edge.colour;
                    }
                }
                else if (edge.successor == -1) {
                    new_state = NODE_BOTTOM;
                    color = 1 - parity_type;
                }
                else if (edge.successor == -2) {
                    new_state = NODE_TOP;
                    color = parity_type;
                }
                else {
                    throw std::invalid_argument("invalid successor for automaton");
                }
                successors[local_state].push_back(new_state);
                successors[local_state].push_back(color);
            }
        }

        new_successors.notify_all();
    }
}

Edge Automaton::getSuccessor(node_id_t local_state, letter_t letter) {
    add_successors(local_state);
    /*
    std::unique_lock<std::mutex> lock(successors_mutex);
    while(local_state >= successors.size() || successors[local_state].empty()) {
        new_successors.wait(lock);
    }
    */
    if (node_type == NodeType::WEAK) {
        const node_id_t state = successors[local_state][letter];
        if (state == NODE_BOTTOM) {
            return Edge(NODE_BOTTOM, 1);
        }
        else if (state == NODE_TOP) {
            return Edge(NODE_TOP, 0);
        }
        else {
            return Edge(state, default_color);
        }
    }
    else {
        const node_id_t state = successors[local_state][2*letter];
        const color_t color = successors[local_state][2*letter+1];
        return Edge(state, color);
    }
}

void Automaton::print_type() const {
    switch (node_type) {
        case NodeType::WEAK:
            if (default_color == 0) {
                std::cout << "Safety";
            }
            else {
                std::cout << "co-Safety";
            }
            break;
        case NodeType::BUCHI:
            std::cout << "Büchi";
            break;
        case NodeType::CO_BUCHI:
            std::cout << "co-Büchi";
            break;
        case NodeType::PARITY:
            std::cout << "Parity (" << getMaxColor() << ")";
            break;
    }
}

void Automaton::print_memory_usage() const {
    size_t cache_size = 0;
    for (const auto& succs : successors) {
        cache_size += succs.size() * sizeof(node_id_t);
    }
    std::cout << "Automaton successors: " << (cache_size / 1024) << std::endl;
}

}
