#include "ParityAutomatonTree.h"

#include <iostream>
#include <iomanip>
#include <algorithm>

namespace aut {

ParityAutomatonTreeBiconditionalNode::ParityAutomatonTreeBiconditionalNode(
        const NodeType node_type, const Parity parity_type, const color_t max_color,
        const node_id_t round_robin_size, const bool parity_child,
        const color_t parity_child_max_color, const int round_robin_child_index, const int parity_child_index,
        std::vector<std::unique_ptr<ParityAutomatonTree>> _children
) :
    ParityAutomatonTreeNode(owl::BICONDITIONAL, node_type, parity_type, max_color, round_robin_size, parity_child, parity_child_max_color, std::move(_children)),
    round_robin_child_index(round_robin_child_index), parity_child_index(parity_child_index)
{ }

ParityAutomatonTreeBiconditionalNode::~ParityAutomatonTreeBiconditionalNode() { }

void ParityAutomatonTreeBiconditionalNode::getInitialState(product_state_t& state) {
    // save index in product state
    state_index = state.size();
    if (round_robin_size > 1) {
        // inverse minimal parity seen
        state.push_back(0);
    }
    for (auto &child : children) {
        child->getInitialState(state);
    }
}

color_t ParityAutomatonTreeBiconditionalNode::getSuccessor(const product_state_t& state, product_state_t& new_state, letter_t letter) {
    if (isBottomState(state)) {
        setBottomState(new_state);
        return 1 - parity_type;
    }
    else if (isTopState(state)) {
        setTopState(new_state);
        return parity_type;
    }

    color_t min_parity = parity_child_max_color;
    if (round_robin_size > 1) {
        min_parity -= state[state_index];
    }

    size_t active_children = 0;

    bool bottom = false;
    bool top = false;
    color_t child_colors[2];

    int child_index = 0;
    for (auto &child : children) {
        color_t child_color = child->getSuccessor(state, new_state, letter);

        child_colors[child_index] = child_color;

        if (child->isBottomState(new_state)) {
            bottom = true;
        }
        else if(child->isTopState(new_state)) {
            top = true;
        }
        else {
            active_children++;
        }

        if (child_index == parity_child_index) {
             min_parity = std::min(min_parity, child_color);
        }

        child_index++;
    }

    if (active_children == 0) {
        if (bottom && top) {
            setBottomState(new_state);
            return 1 - parity_type;
        }
        else {
            setTopState(new_state);
            return parity_type;
        }
    }
    if (round_robin_size == 0) {
        // only safety/reachability/weak children
        if (child_colors[0] == child_colors[1]) {
            return parity_type;
        }
        else {
            return 1 - parity_type;
        }
    }
    else {
        color_t color;
        if (child_colors[round_robin_child_index] == 0) {
            color = min_parity;
            min_parity = parity_child_max_color;
        }
        else {
            color = child_colors[parity_child_index] + 1;
        }

        if (round_robin_size > 1) {
            new_state[state_index] = parity_child_max_color - min_parity;
        }
        return color;
    }
}

void ParityAutomatonTreeBiconditionalNode::setState(product_state_t& new_state, node_id_t state) {
    if (round_robin_size > 1) {
        new_state[state_index] = state;
    }
    children[0]->setState(new_state, state);
    children[1]->setState(new_state, state);
}

void ParityAutomatonTreeBiconditionalNode::setTopState(product_state_t& new_state) {
    if (round_robin_size > 1) {
        setState(new_state, NODE_NONE);
        new_state[state_index] = NODE_TOP;
    }
    else {
        // decide to set both to top
        children[0]->setTopState(new_state);
        children[1]->setTopState(new_state);
    }
}

void ParityAutomatonTreeBiconditionalNode::setBottomState(product_state_t& new_state) {
    if (round_robin_size > 1) {
        setState(new_state, NODE_NONE);
        new_state[state_index] = NODE_BOTTOM;
    }
    else {
        // decide to set first to bottom and second to top
        children[0]->setBottomState(new_state);
        children[1]->setTopState(new_state);
    }
}

bool ParityAutomatonTreeBiconditionalNode::isTopState(const product_state_t& state) const {
    if (round_robin_size > 1) {
        return state[state_index] == NODE_TOP;
    }
    else {
        return children[0]->isTopState(state) && children[1]->isTopState(state);
    }
}
bool ParityAutomatonTreeBiconditionalNode::isBottomState(const product_state_t& state) const {
    if (round_robin_size > 1) {
        return state[state_index] == NODE_BOTTOM;
    }
    else {
        return children[0]->isBottomState(state) && children[1]->isTopState(state);
    }
}

}
