#include "ParityAutomatonTree.h"

#include <iostream>
#include <iomanip>
#include <algorithm>

namespace aut {

bool cmp_trees(const std::unique_ptr<ParityAutomatonTree>& t1, const std::unique_ptr<ParityAutomatonTree>& t2) {
    if (t1->node_type < t2->node_type) {
        return true;
    }
    else if (t1->node_type > t2->node_type) {
        return false;
    }
    else {
        const letter_t size1 = t1->getMaximumAlphabetSize();
        const letter_t size2 = t2->getMaximumAlphabetSize();
        if (size1 < size2) {
            return true;
        }
        else if (size1 > size2) {
            return false;
        }
        else {
            const auto alphabet1 = t1->getAlphabet();
            const auto alphabet2 = t2->getAlphabet();
            if (alphabet1 < alphabet2) {
                return true;
            }
            else if (alphabet1 > alphabet2) {
                return false;
            }
            else {
                return t1->getMinIndex() < t2->getMinIndex();
            }
        }
    }
}

AutomatonTreeStructure::AutomatonTreeStructure(owl::EmersonLeiAutomaton automaton) {
    for (auto a : automaton.automata()) {
        automata.emplace_back(std::move(a));
    }
    const std::unique_ptr<owl::LabelledTree<owl::Tag, owl::Reference>> structure = automaton.structure();
    tree = constructTree(structure);
}

std::unique_ptr<ParityAutomatonTree> AutomatonTreeStructure::constructTree(const std::unique_ptr<owl::LabelledTree<owl::Tag, owl::Reference>>& tree) {
    if (tree->is_leaf()) {
        const owl::Reference& reference = tree->label2();
        Automaton& automaton = automata[reference.index];
        std::unique_ptr<ParityAutomatonTree> leaf(new ParityAutomatonTreeLeaf(automaton, std::move(reference)));
        return leaf;
    }
    else {
        const owl::Tag tag = tree->label1();

        NodeType node_type = NodeType::WEAK;
        Parity parity_type = Parity::EVEN;
        color_t max_color = 1;
        bool parity_child = false;
        color_t parity_child_max_color = 0;
        Parity parity_child_parity_type = Parity::EVEN;
        node_id_t round_robin_size = 0;
        int round_robin_child_index = -1;
        int parity_child_index = -1;

        std::vector<std::unique_ptr<ParityAutomatonTree>> children;
        for (const std::unique_ptr<owl::LabelledTree<owl::Tag, owl::Reference>>& child : tree->children()) {
            std::unique_ptr<ParityAutomatonTree> pchild = constructTree(child);
            if (pchild->node_type == NodeType::PARITY) {
                if (parity_child) {
                    throw std::invalid_argument("unsupported automaton tree");
                }
                else {
                    parity_child = true;
                    parity_child_parity_type = pchild->parity_type;
                    parity_child_max_color = pchild->max_color;
                }
                if (tag == owl::BICONDITIONAL) {
                    round_robin_size++;
                }
            }
            else if (pchild->node_type == NodeType::BUCHI) {
                if (tag == owl::CONJUNCTION || tag == owl::BICONDITIONAL) {
                    round_robin_size++;
                }
            }
            else if (pchild->node_type == NodeType::CO_BUCHI) {
                if (tag == owl::DISJUNCTION || tag == owl::BICONDITIONAL) {
                    round_robin_size++;
                }
            }

            node_type = join_node_type(node_type, pchild->node_type);

            children.push_back(std::move(pchild));
        }

        // sort children from "small,simple" to "large,complex" as a heuristic for more efficient product construction
        std::sort(children.begin(), children.end(), cmp_trees);

        if (tag == owl::BICONDITIONAL) {
            NodeType t1 = children[0]->node_type;
            NodeType t2 = children[1]->node_type;
            node_type = join_node_type_biconditional(t1, t2);

            if (round_robin_size > 0) {
                // need to treat one child as a parity child and other as a buchi/co-buchi child
                parity_child = true;
                if (t1 == NodeType::BUCHI || t1 == NodeType::CO_BUCHI) {
                    round_robin_child_index = 0;
                    parity_child_index = 1;
                }
                else if (t2 == NodeType::BUCHI || t2 == NodeType::CO_BUCHI) {
                    round_robin_child_index = 1;
                    parity_child_index = 0;
                }
                else if (t1 == NodeType::PARITY) {
                    round_robin_child_index = 1;
                    parity_child_index = 0;
                }
                else if (t2 == NodeType::PARITY) {
                    round_robin_child_index = 0;
                    parity_child_index = 1;
                }
            }
        }

        if (node_type == NodeType::PARITY) {
            if (tag == owl::CONJUNCTION || tag == owl::DISJUNCTION) {
                if (tag == owl::CONJUNCTION) {
                    parity_type = Parity::ODD;
                }
                else {
                    parity_type = Parity::EVEN;
                }
                if (parity_child) {
                    if (parity_type != parity_child_parity_type) {
                        parity_child_max_color++;
                    }
                    max_color = parity_child_max_color;
                    if (round_robin_size > 0 && max_color % 2 != 0) {
                        max_color++;
                    }
                }
                else {
                    // needs to have co-buchi and buchi children
                    max_color = 2;
                }
            }
            else {
                if (round_robin_size > 0) {
                    if (round_robin_size > 0 && children[round_robin_child_index]->node_type == NodeType::CO_BUCHI) {
                        parity_type = (Parity)(1 - (int)children[parity_child_index]->parity_type);
                    }
                    else {
                        parity_type = children[parity_child_index]->parity_type;
                    }
                    max_color = children[parity_child_index]->max_color + 1;
                    parity_child_max_color = children[parity_child_index]->max_color;
                }
                else {
                    parity_type = children[parity_child_index]->parity_type;
                    max_color = parity_child_max_color;
                }
            }
        }
        else if (node_type == NodeType::BUCHI) {
            parity_type = Parity::EVEN;
        }
        else if (node_type == NodeType::CO_BUCHI) {
            parity_type = Parity::ODD;
        }

        if (tag == owl::CONJUNCTION || tag == owl::DISJUNCTION) {
            return std::unique_ptr<ParityAutomatonTree>(
                    new ParityAutomatonTreeNode(
                        tag, node_type, parity_type, max_color, round_robin_size, parity_child, parity_child_max_color, std::move(children)));
        }
        else {
            return std::unique_ptr<ParityAutomatonTree>(
                    new ParityAutomatonTreeBiconditionalNode(
                        node_type, parity_type, max_color, round_robin_size, parity_child, parity_child_max_color,
                        round_robin_child_index, parity_child_index, std::move(children)));
        }
    }
}

Parity AutomatonTreeStructure::getParityType() const {
    return tree->parity_type;
}

color_t AutomatonTreeStructure::getMaxColor() const {
    return tree->max_color;
}

void AutomatonTreeStructure::getInitialState(product_state_t& state) {
    tree->getInitialState(state);
}

color_t AutomatonTreeStructure::getSuccessor(const product_state_t& state, product_state_t& new_state, letter_t letter) {
    return tree->getSuccessor(state, new_state, letter);
}

bool AutomatonTreeStructure::isTopState(const product_state_t& state) const {
    return tree->isTopState(state);
}

bool AutomatonTreeStructure::isBottomState(const product_state_t& state) const {
    return tree->isBottomState(state);
}

void AutomatonTreeStructure::clear() {
    automata.clear();
    // free memory from automata and tree
    std::deque<Automaton>().swap(automata);
    tree.reset();
}

void AutomatonTreeStructure::print(const bool print_formula) const {
    int i = 0;
    for (const auto& automaton : automata) {
        std::cout << "A[" << i << "]: ";
        automaton.print_type();
        std::cout << std::endl;
        i++;
    }
    tree->print(print_formula);
}

void AutomatonTreeStructure::print_memory_usage() const {
    for (const auto& automaton : automata) {
        automaton.print_memory_usage();
    }
}

}
