#include "PGArena.h"

#include <iostream>
#include <iomanip>
#include <sstream>
#include <cmath>
#include <limits>
#include <algorithm>
#include <cstdint>
#include <bitset>
#include <unordered_map>
#include <queue>
#include <memory>

#include <boost/functional/hash.hpp>

namespace pg {

PGArena::PGArena(const size_t n_inputs, const size_t n_outputs, aut::AutomatonTreeStructure _structure) :
    structure(std::move(_structure)),
    n_inputs(n_inputs),
    n_outputs(n_outputs),
    complete(false),
    solved(false),
    parity_type(structure.getParityType()),
    n_colors(structure.getMaxColor() + 1),
    initial_node(0),
    n_env_actions(1u << n_inputs),
    n_sys_actions(1u << n_outputs),
    n_env_nodes(0),
    n_sys_edges(0)
{
    // reserve some space in the vector to avoid initial resizing, which needs locking for the parallel construction
    constexpr size_t RESERVE = 4096;
    sys_succs_begin.reserve(RESERVE);
    sys_succs.reserve(RESERVE);
    sys_output.reserve(RESERVE);
}

PGArena::~PGArena() { }

template <typename Container>
struct container_hash {
    std::size_t operator()(Container const& c) const {
        return boost::hash_range(c.cbegin(), c.cend());
    }
};

int PGArena::computeStateLabels(std::vector<node_id_t>& visited_map, std::vector<int>& accumulated_bits) {
    // get visited states
    std::vector<std::set<node_id_t>> visited_states(product_state_size);
    for (node_id_t i = 0; i < n_env_nodes; i++) {
        if (visited_map[i] != NODE_NONE) {
            for (size_t j = 0; j < product_state_size; j++) {
                // none states are don't cares
                if (product_states[i][j] != NODE_NONE) {
                    visited_states[j].insert(product_states[i][j]);
                }
            }
        }
    }
    // get map for visited states
    std::vector<std::map<node_id_t, node_id_t>> inner_state_map(product_state_size);
    for (size_t i = 0; i < product_state_size; i++) {
        node_id_t new_id = 0;
        for (const auto& old_id : visited_states[i]) {
            inner_state_map[i].insert({ old_id, new_id });
            new_id++;
        }
    }

    accumulated_bits.resize(product_state_size + 1);
    // get number of bits
    int state_label_bits = 0;
    for (size_t i = 0; i < product_state_size; i++) {
        accumulated_bits[i] = state_label_bits;
        size_t state_size = visited_states[i].size();
        if (state_size > 1) {
            state_label_bits += log2(state_size - 1) + 1;
        }
    }
    accumulated_bits[product_state_size] = state_label_bits;
    constexpr int node_id_t_bitwidth = std::numeric_limits<node_id_t>::digits;
    if (state_label_bits >= node_id_t_bitwidth) {
        return -1;
    }

    // build state ids
    state_labels.resize(n_env_nodes);
    for (node_id_t i = 0; i < n_env_nodes; i++) {
        if (visited_map[i] != NODE_NONE) {
            node_id_t id_number = 0;
            node_id_t id_dontcare = 0;
            for (size_t j = 0; j < product_state_size; j++) {
                if (product_states[i][j] == NODE_NONE) {
                    id_dontcare |= (((~0u) << accumulated_bits[j]) & ((1u << accumulated_bits[j+1]) - 1));
                }
                else {
                    id_number |= (inner_state_map[j][product_states[i][j]] << accumulated_bits[j]);
                }
            }
            state_labels[i] = SpecSeq<node_id_t>(id_number, id_dontcare);
        }
    }

    return state_label_bits;
}

void PGArena::constructArena(const bool parallel, const bool print_product_states, const bool print_product_successors) {
    product_state_t initial_state;
    structure.getInitialState(initial_state);
    product_state_size = initial_state.size();
    bool has_lock = false;

    std::queue<product_state_t> queue;
    std::unordered_map< product_state_t, node_id_t, container_hash<product_state_t> > state_map;

    state_map.insert({initial_state, initial_node});
    queue.push(std::move(initial_state));
    sys_succs_begin.push_back(0);

    node_id_t max_env_node = 0;

    while (!solved && !queue.empty()) {

        const product_state_t state = std::move(queue.front());
        queue.pop();

        if (print_product_states) {
            std::cout << "Computing successors for " << std::setw(4) << n_env_nodes << " = (";
            for (const auto s : state) {
                if (s == NODE_TOP) {
                    std::cout << "  ⊤";
                }
                else if (s == NODE_BOTTOM) {
                    std::cout << "  ⊥";
                }
                else if (s == NODE_NONE) {
                    std::cout << "  -";
                }
                else {
                    std:: cout << " " << std::setw(2) << s;
                }
            }
            std::cout << " )";
            if (print_product_successors) {
                std::cout << " : [";
            }
            else {
                std::cout << std::endl;
            }
        }

        edge_id_t cur_n_sys_edges = 0;

        for (letter_t i = 0; i < n_env_actions; i++) {

            // new sys node
            for (letter_t o = 0; o < n_sys_actions; o++) {
                letter_t letter = (o << n_inputs) + i;

                product_state_t new_state(product_state_size);
                color_t color = structure.getSuccessor(state, new_state, letter);

                if (!structure.isBottomState(new_state)) {
                    node_id_t succ;

                    if (structure.isTopState(new_state)) {
                        succ = NODE_TOP;
                    }
                    else {
                        auto it = state_map.find(new_state);
                        if (it == state_map.end()) {
                            succ = max_env_node + 1;
                            max_env_node++;
                            state_map.insert({new_state, succ});
                            queue.push(std::move(new_state));
                        }
                        else {
                            succ = it->second;
                        }
                    }

                    if (print_product_successors) {
                        if (succ == NODE_TOP) {
                            std::cout << "   ⊤   ";
                        }
                        else {
                            std::cout << "  " << std::setw(2) << succ;
                            std::cout << ":" << std::setw(2) << color;
                        }
                    }

                    sys_output.push_back(o);
                    if (parallel && (sys_succs.size() == sys_succs.capacity())) {
                        arena_mutex.lock();
                        has_lock = true;
                    }
                    sys_succs.push_back(Edge(succ, color));
                    cur_n_sys_edges++;
                    if (has_lock) {
                        arena_mutex.unlock();
                        has_lock = false;
                    }
                }
                else {
                    if (print_product_successors) {
                        std::cout << "   ⊥   ";
                    }
                }
            }

            if (parallel && (sys_succs_begin.size() == sys_succs_begin.capacity())) {
                arena_mutex.lock();
                has_lock = true;
            }
            sys_succs_begin.push_back(sys_succs.size());
            if (has_lock) {
                arena_mutex.unlock();
                has_lock = false;
            }
        }
        if (print_product_successors) {
            std::cout << "]" << std::endl;
        }

        if (parallel) {
            size_mutex.lock();
        }
        n_sys_edges += cur_n_sys_edges;
        n_env_nodes++;
        if (parallel) {
            size_mutex.unlock();
            change.notify_all();
        }
    }

    if (parallel) {
        size_mutex.lock();
    }
    assert(n_env_nodes * n_env_actions + 1 == sys_succs_begin.size());
    assert(n_sys_edges == sys_succs.size());
    complete = true;
    if (parallel) {
        size_mutex.unlock();
        change.notify_all();
    }

    // fill vector for product states
    product_states.resize(n_env_nodes);
    for (auto& entry : state_map) {
        if (entry.second < n_env_nodes) {
            product_states[entry.second] = std::move(entry.first);
        }
    }

    // clear structure, not needed any more
    structure.clear();
}

void PGArena::print_basic_info() const {
    std::cout << "Number of env nodes  : " << n_env_nodes << std::endl;
    std::cout << "Number of env actions: " << n_env_actions << std::endl;
    std::cout << "Number of sys edges  : " << n_sys_edges << std::endl;
    std::cout << "Number of colors     : " << n_colors << std::endl;
}

void PGArena::print(std::ostream& out, Player winner) const {
    size_t n_sys_nodes = n_env_nodes * n_env_actions;
    size_t total_nodes = n_env_nodes + n_sys_nodes + n_sys_edges + 3;
    size_t unexplored_node = total_nodes - 1;
    size_t bottom_node = total_nodes - 2;
    size_t top_node = total_nodes - 3;
    color_t max_color = n_colors - 1;
    if (max_color % 2 != 0) {
        max_color++;
    }
    color_t neutral_color = 0;

    size_t id_size = 1 + (size_t)log10(total_nodes);
    size_t color_size = 1 + (size_t)log10(max_color);

    /*
    out << "min ";
    switch(parity_type) {
        case Parity::ODD:
            out << "odd ";
            break;
        case Parity::EVEN:
            out << "even ";
            break;
    }
    */
    out << "parity " <<  + total_nodes << ";" << std::endl;

    // environment nodes
    for (node_id_t i = 0; i < n_env_nodes; i++) {
        out << std::setw(id_size) << i << " "
            << std::setw(color_size) << neutral_color << " "
            << (1 - parity_type) << " ";
        const node_id_t succs_begin = i * n_env_actions;
        const node_id_t succs_end = succs_begin + n_env_actions;
        for (node_id_t j = succs_begin; j != succs_end; j++) {
            out << std::setw(id_size) << (j + n_env_nodes);
            if (j + 1 != succs_end) {
                out << ",";
            }
        }
        out << ";" << std::endl;
    }
    out << std::endl;

    // system nodes
    for (node_id_t env_node = 0; env_node < n_env_nodes; env_node++) {
        for (node_id_t input = 0; input < n_env_actions; input++) {
            size_t i = env_node * n_env_actions + input;

            out << std::setw(id_size) << (i + n_env_nodes) << " "
                << std::setw(color_size) << neutral_color << " "
                << parity_type << " ";
            if (getSysSuccsBegin(i) == getSysSuccsEnd(i)) {
                out << std::setw(id_size) << bottom_node;
            }
            else {
                for (edge_id_t j = getSysSuccsBegin(i); j != getSysSuccsEnd(i); j++) {
                    out << std::setw(id_size) << (j + n_env_nodes + n_sys_nodes);
                    if (j + 1 != getSysSuccsEnd(i)) {
                        out << ",";
                    }
                }
            }
            SpecSeq<letter_t> ti_input(input);
            out << " \"" << ti_input.toString(n_inputs) << "\";" << std::endl;

            for (edge_id_t j = getSysSuccsBegin(i); j != getSysSuccsEnd(i); j++) {
                const Edge edge = getSysEdge(j);
                out << std::setw(id_size) << (j + n_env_nodes + n_sys_nodes) << " "
                    << std::setw(color_size) << (max_color - edge.color) << " "
                    << parity_type << " ";
                if (edge.successor == NODE_TOP) {
                    out << std::setw(id_size) << top_node;
                }
                else if (edge.successor >= n_env_nodes) {
                    out << std::setw(id_size) << unexplored_node;
                }
                else {
                    out << std::setw(id_size) << edge.successor;
                }
                SpecSeq<letter_t> ti_output(getSysOutput(j));
                out << " \"" << ti_output.toString(n_outputs) << "\";" << std::endl;
            }
        }
    }
    out << std::endl;

    // top, bottom and unexplored nodes
    out << std::setw(id_size) << top_node << " "
        << std::setw(color_size) << parity_type << " " << parity_type << " "
        << top_node << " \"top\";" << std::endl;
    out << std::setw(id_size) << bottom_node << " "
        << std::setw(color_size) << (1 - parity_type) << " " << (1 - parity_type) << " "
        << bottom_node << " \"bottom\";" << std::endl;
    out << std::setw(id_size) << unexplored_node << " ";
    if (winner == SYS_PLAYER) {
        out << std::setw(color_size) << (1 - parity_type) << " " << (1 - parity_type) << " ";
    }
    else if (winner == ENV_PLAYER) {
        out << std::setw(color_size) << parity_type << " " << parity_type << " ";
    }
    else {
        out << std::setw(color_size) << "-" << " - ";
    }
    out << unexplored_node << " \"unexplored\";" << std::endl;
}

void PGArena::print_memory_usage() const {
    std::cout << "Sys output: " << (sys_output.size() * sizeof(letter_t) / 1024) << std::endl;
    std::cout << "Sys succs begin: " << (sys_succs_begin.size() * sizeof(edge_id_t) / 1024) << std::endl;
    std::cout << "Sys succs: " << (sys_succs_begin.size() * sizeof(Edge) / 1024) << std::endl;
    structure.print_memory_usage();
}

}
