#pragma once

#include <atomic>
#include <mutex>
#include <condition_variable>

#include "Definitions.h"
#include "util/Quine.h"
#include "aut/ParityAutomatonTree.h"

namespace pg {

class PGArena {
private:
    aut::AutomatonTreeStructure structure;

    std::vector<letter_t> sys_output;
    std::vector<edge_id_t> sys_succs_begin;
    std::vector<Edge> sys_succs;

    size_t product_state_size;
    std::vector<product_state_t> product_states;
    std::vector<SpecSeq<node_id_t>> state_labels;

public:
    const size_t n_inputs;
    const size_t n_outputs;

    PGArena(const size_t n_inputs, const size_t n_outputs, aut::AutomatonTreeStructure _structure);
    ~PGArena();

    void constructArena(const bool parallel = false, const bool print_product_states = false, const bool print_product_successors = false);
    int computeStateLabels(std::vector<node_id_t>& visited_map, std::vector<int>& accumulated_bits);

    // mutex for reading from or resizing the arena vectors
    std::mutex arena_mutex;

    // mutex for accessing size of the arena
    std::mutex size_mutex;

    // condition variable signalling change in size of the arena
    std::condition_variable change;

    // variable signalling that the arena is completely constructed
    std::atomic<bool> complete;

    // variable signalling that the arena is already solved
    std::atomic<bool> solved;

    const Parity parity_type;
    const color_t n_colors;

    const node_id_t initial_node;

    const letter_t n_env_actions;
    const letter_t n_sys_actions;

    std::atomic<node_id_t> n_env_nodes;
    std::atomic<edge_id_t> n_sys_edges;

    void print(std::ostream& out, Player winner = UNKNOWN) const;
    void print_basic_info() const;
    void print_memory_usage() const;

    inline Edge getSysEdge(edge_id_t sys_edge) const { return sys_succs[sys_edge]; }
    inline letter_t getSysOutput(edge_id_t sys_edge) const { return sys_output[sys_edge]; }
    inline edge_id_t getSysSuccsBegin(node_id_t sys_node) const { return sys_succs_begin[sys_node]; }
    inline edge_id_t getSysSuccsEnd(node_id_t sys_node) const { return sys_succs_begin[sys_node + 1]; }

    inline node_id_t getEnvSuccsBegin(node_id_t env_node) const { return env_node << n_inputs; };
    inline node_id_t getEnvSuccsEnd(node_id_t env_node) const { return (env_node + 1) << n_inputs; };

    inline SpecSeq<node_id_t> getStateLabel(node_id_t env_node) const {
        return state_labels[env_node];
    };
};

}
