#include "PGSolver.h"

#include <bitset>
#include <deque>
#include <set>
#include <mutex>

#include "omp.h"

namespace pg {

PGSolver::PGSolver(pg::PGArena& arena, const bool onthefly_construction, const int num_threads, const bool compact_colors, const bool verbose) :
    arena(arena),
    onthefly_construction(onthefly_construction),
    num_threads(num_threads),
    compact_colors(compact_colors),
    verbose(verbose),
    winner(UNKNOWN),
    n_colors(arena.n_colors)
{
    init_solver();
}

PGSolver::~PGSolver() { }

void PGSolver::init_solver() {
    n_env_nodes = arena.n_env_nodes;
    n_sys_edges = arena.n_sys_edges;
    n_sys_nodes = n_env_nodes * arena.n_env_actions;
}

void PGSolver::preprocess_and_solve_game() {
    if (compact_colors) {
        reduce_colors();
    }
    else {
        copy_colors();
    }
    solve_game();
}


void PGSolver::reduce_colors() {
    n_colors = arena.n_colors;

    std::vector<edge_id_t> color_count(n_colors, 0);

    for (edge_id_t i = 0; i < n_sys_edges; i++) {
        const color_t c = arena.getSysEdge(i).color;
        color_count[c]++;
    }

    color_map.resize(n_colors);

    color_t cur_color = 0;
    for (color_t c = 0; c < n_colors; c++) {
        if (color_count[c] != 0) {
            if ((c % 2) != (cur_color % 2)) {
                cur_color++;
            }
            color_map[c] = cur_color;
        }
    }

    n_colors = cur_color + 1;
}

void PGSolver::copy_colors() {
    n_colors = arena.n_colors;
    color_map.resize(n_colors);
    for (color_t c = 0; c < n_colors; c++) {
        color_map[c] = c;
    }
}

void PGSolver::solve() {
    // set number of threads for parallel solving
    int max_threads;
    if (num_threads > 0) {
        max_threads = num_threads;
    }
    else {
        max_threads = omp_get_max_threads();
    }
    if (onthefly_construction && max_threads > 1) {
        // leave one thread for construction of arena
        max_threads--;
    }
    omp_set_num_threads(max_threads);
    parallel = max_threads > 1;

    if (onthefly_construction) {
        while (!arena.solved) {
            {
                std::unique_lock<std::mutex> lock(arena.size_mutex);
                while (!arena.complete && arena.n_env_nodes == n_env_nodes) {
                    arena.change.wait(lock);
                }

                init_solver();
            }

            std::unique_lock<std::mutex> lock(arena.arena_mutex);
            preprocess_and_solve_game();

            if (winner != UNKNOWN) {
                arena.solved = true;
            }
        }
    }
    else {
        init_solver();
        preprocess_and_solve_game();
    }
}

void PGSolver::print_debug(const std::string& str) const {
    if (verbose) {
        std::cout << str << std::endl;
    }
}

Player PGSolver::getWinner() const {
    return winner;
}

bool PGSolver::constructMealyMachine(mealy::MealyMachine& m, const bool add_product_labels) const {
    mealy::machine_t machine;

    const SpecSeq<letter_t> any_output = true_clause<letter_t>(arena.n_outputs);
    bool has_top_state = false;

    std::vector<node_id_t> state_map(n_env_nodes, NODE_NONE);

    std::deque<node_id_t> queue;
    queue.push_back(arena.initial_node);
    machine.push_back({});

    state_map[arena.initial_node] = 0;

    while (!queue.empty()) {
        node_id_t env_node = queue.front();
        queue.pop_front();

        mealy::state_id_t state = state_map[env_node];
        node_id_t sys_node = env_node * arena.n_env_actions;

        std::map< mealy::Successor, std::vector<letter_t>> input_list;

        for (letter_t input = 0; input < arena.n_env_actions; input++, sys_node++) {
            std::map<node_id_t, std::vector<letter_t>> successor_list;

            for (edge_id_t j = arena.getSysSuccsBegin(sys_node); j != arena.getSysSuccsEnd(sys_node); j++) {
                if (sys_successors[j]) {
                    const Edge edge = arena.getSysEdge(j);
                    letter_t output = arena.getSysOutput(j);

                    successor_list[edge.successor].push_back(output);
                }
            }

            // Heuristic: if possible, choose successor that was already explored,
            // and among those, choose successor with with maximum non-determinism

            auto succ_it = successor_list.begin();
            bool succ_explored = false;
            size_t succ_num_outputs = 0;

            for (auto it = successor_list.begin(); it != successor_list.end(); it++) {
                const node_id_t cur_successor = it->first;
                if (cur_successor == NODE_TOP) {
                    // always choose top node
                    succ_it = it;
                    break;
                }
                else {
                    bool cur_succ_explored = (state_map[cur_successor] != NODE_NONE);
                    size_t cur_succ_num_outputs = it->second.size();
                    if (
                            (cur_succ_explored && !succ_explored) ||
                            (cur_succ_explored == succ_explored && cur_succ_num_outputs > succ_num_outputs)
                    ) {
                        succ_it = it;
                        succ_num_outputs = cur_succ_num_outputs;
                        succ_explored = cur_succ_explored;
                    }
                }
            }

            const node_id_t successor = succ_it->first;
            const std::vector<letter_t>& outputs = succ_it->second;

            mealy::state_id_t mealy_successor;
            if (successor == NODE_TOP) {
                mealy_successor = mealy::TOP_STATE;
            }
            else {
                if (state_map[successor] == NODE_NONE) {
                    mealy::state_id_t successor_state = machine.size();
                    machine.push_back({});
                    state_map[successor] = successor_state;
                    queue.push_back(successor);
                }
                mealy_successor = state_map[successor];
            }

            SpecSeq<letter_t> min_output = quine_minimal_implicant(arena.n_outputs, outputs);

            if (mealy_successor != mealy::TOP_STATE || min_output != any_output) {
                if (mealy_successor == mealy::TOP_STATE) {
                    has_top_state = true;
                }
                input_list[mealy::Successor(mealy_successor, min_output)].push_back(input);
            }
        }

        for (const auto& entry : input_list) {
            std::vector<SpecSeq<letter_t>> inputs = quine_prime_implicants(arena.n_inputs, entry.second);
            for (const auto& input : inputs) {
                machine[state].push_back( mealy::Transition(entry.first.successor, input, entry.first.output) );
            }
        }
    }

    mealy::state_id_t top_state;
    if (has_top_state) {
        top_state = machine.size();
        for (mealy::state_id_t s = 0; s < machine.size(); s++) {
            for (auto& t : machine[s]) {
                if (t.nextState == mealy::TOP_STATE) {
                    t.nextState = top_state;
                }
            }
        }

        machine.push_back({});
    }

    mealy::state_id_t numStates = machine.size();
    m.setMachine(std::move(machine));

    if (add_product_labels) {
        std::vector<int> accumulated_bits;
        int state_label_bits = arena.computeStateLabels(state_map, accumulated_bits);
        if (state_label_bits < 0) {
            return false;
        }
        else {
            std::vector<SpecSeq<node_id_t>> labels(numStates);
            for (node_id_t i = 0; i < n_env_nodes; i++) {
                if (state_map[i] != NODE_NONE) {
                    labels[state_map[i]] = arena.getStateLabel(i);
                }
            }
            if (has_top_state) {
                labels[top_state] = true_clause<node_id_t>(state_label_bits);
            }
            m.setStateLabels(std::move(labels), state_label_bits, std::move(accumulated_bits));

            return true;
        }
    }
    else {
        return false;
    }
}

}
