#include "OptionParser.h"

#include <iostream>
#include <boost/algorithm/string/predicate.hpp>

namespace strix {

OptionParser::OptionParser(const int argc, const char* argv[]) :
    argc(argc),
    argv(argv),
    input_options("Input options"),
    output_options("Output options"),
    synthesis_options("Synthesis options"),
    finetuning_options("Fine-tuning options"),
    debugoutput_options("Debug output options"),
    misc_options("Miscellaneous options"),
    hidden_options(),
    positional_options(),
    options()
{
    input_options.add_options()
        ("ltl", "read input file as single LTL formula instead of TLSF")
        ("formula,f", po::value<std::string>(), "use the LTL formula given by arg instead of FILE")
        ("ins", po::value<std::string>(), "use this comma-separated list of input propositions")
        ("outs", po::value<std::string>(), "use this comma-separated list of output propositions")
    ;
    output_options.add_options()
        ("kiss,k", "output Mealy machine in KISS format")
        ("binary", "output AIGER circuit in binary format")
        ("dot,d", "output Mealy machine in Graphviz DOT format")
        ("bdd,b", "output BDDs in DOT format")
        ("parity-game,p", "output parity game in PGSolver format")
        ("output,o", po::value<std::string>(), "output controller to this file")
    ;
    synthesis_options.add_options()
        ("realizability,r", "only check realizability")
        ("labels,l", "use product state labels for Mealy machine")
        ("minimize,m", "minimize Mealy machine")
        ("auto,a", "automatically find configuration producing minimal AIGER circuit")
    ;
    finetuning_options.add_options()
        ("monolithic", "do not use formula splitting")
        ("splitting", po::value<std::string>(), "safety formula splitting: never, auto, always")
        ("no-onthefly", "do not construct and solve arena in on-the-fly")
        ("no-simplify-formula", "do not simplify the formula")
        ("threads", po::value<int>(), "set the number of solver threads (0=auto)")
        ("no-compact-colors", "do not compact the colors by removing unused colors")
        ("validate-jni", "validate JNI interface")
        ("optimize-heap", "optimize Java heap size")
        ("heap-size", po::value<int>(), "maximum heap size in GB (0=auto)")
    ;
    debugoutput_options.add_options()
        ("print-arena-info", "print basic arena information")
        ("print-formula", "print formulas")
        ("print-tree", "print automaton tree")
        ("print-product-states", "print states of product construction")
        ("print-product-successors", "print successors of product construction")
        ("verbose-solver", "verbose output for solver")
    ;
    misc_options.add_options()
        ("classpath", po::value<std::string>(), "location of Owl library")
        ("timing,t", "measure and print timing information")
        ("help,h", "display this help text and exit")
    ;
    hidden_options.add_options()
        ("input-file", po::value<std::string>(), "input file")
    ;
    options
        .add(input_options)
        .add(output_options)
        .add(synthesis_options)
        .add(finetuning_options)
        .add(debugoutput_options)
        .add(misc_options)
        .add(hidden_options);
    positional_options.add("input-file", -1);
}

OptionParser::~OptionParser() { }

void OptionParser::print_help() const {
    std::cout << "Usage: " << argv[0] << " [OPTIONS] FILE" << std::endl;
    std::cout << "Synthesize a controller according to the TLSF/LTL specification in the given FILE." << std::endl;
    std::cout << "Outputs the controller in textual AIGER format by default." << std::endl;
    std::cout << std::endl;
    std::cout << input_options << std::endl;
    std::cout << output_options << std::endl;
    std::cout << synthesis_options << std::endl;
    std::cout << finetuning_options << std::endl;
    std::cout << debugoutput_options << std::endl;
    std::cout << misc_options << std::endl;
}

Options OptionParser::parse_options() const {
    po::variables_map vm;

    po::store(po::command_line_parser(argc, argv).
            options(options).
            positional(positional_options).
            run(), vm);
    po::notify(vm);

    Options options;

    // input options
    if (vm.count("input-file")) {
        options.input_file = vm["input-file"].as<std::string>();
    }
    if (vm.count("formula")) {
        options.formula = vm["formula"].as<std::string>();
    }
    options.ltl = !options.formula.empty() || vm.count("ltl") > 0;
    if (vm.count("ins") > 0) {
        std::string inputs_string = vm["ins"].as<std::string>();
        boost::split(options.inputs,inputs_string,boost::is_any_of(" ,"));
    }
    if (vm.count("outs") > 0) {
        std::string outputs_string = vm["outs"].as<std::string>();
        boost::split(options.outputs,outputs_string,boost::is_any_of(" ,"));
    }

    // output options
    if (vm.count("output")) {
        options.output_file = vm["output"].as<std::string>();
    }
    options.mealy = vm.count("kiss") > 0;
    options.binary = vm.count("binary") > 0;
    options.dot = vm.count("dot") > 0;
    options.bdd = vm.count("bdd") > 0;
    options.parity_game = vm.count("parity-game") > 0;

    // synthesis options
    options.realizability = vm.count("realizability") > 0;
    options.labels = vm.count("labels") > 0;
    options.minimize = vm.count("minimize") > 0;
    options.auto_aiger = vm.count("auto") > 0;
    if (options.auto_aiger) {
        // need to minimize and produce labels for this option
        options.labels = true;
        options.minimize = true;
    }

    // fine-tuning options
    options.monolithic = vm.count("monolithic") > 0;
    if (vm.count("splitting")) {
        std::string splitting = vm["splitting"].as<std::string>();
        boost::algorithm::to_lower(splitting);
        if (splitting == "never") {
            options.splitting = owl::NEVER;
        }
        else if (splitting == "auto") {
            options.splitting = owl::AUTO;
        }
        else if (splitting == "always") {
            options.splitting = owl::ALWAYS;
        }
        else {
            throw std::invalid_argument("Invalid splitting: " + vm["splitting"].as<std::string>());
        }
    }
    else {
        options.splitting = owl::AUTO;
    }
    options.onthefly = vm.count("no-onthefly") == 0;
    options.simplify_formula = vm.count("no-simplify-formula") == 0;
    if (vm.count("threads")) {
        options.threads = vm["threads"].as<int>();
        if (options.threads < 0) {
            throw std::invalid_argument("Invalid number of threads: " + options.threads);
        }
    }
    else {
        options.threads = 0;
    }
    options.compact_colors = vm.count("no-compact-colors") == 0;
    options.validate_jni = vm.count("validate-jni") > 0;
    options.aggressive_heap_optimizations = vm.count("optimize-heap") > 0;
    if (vm.count("heap-size")) {
        options.max_heap_size_gb = vm["heap-size"].as<int>();
        if (options.max_heap_size_gb < 0) {
            throw std::invalid_argument("Invalid heap size: " + options.max_heap_size_gb);
        }
    }
    else {
        options.max_heap_size_gb = 0;
    }

    // debug output options
    options.print_arena_info = vm.count("print-arena-info") > 0;
    options.print_formula = vm.count("print-formula") > 0;
    options.print_tree = vm.count("print-tree") > 0;
    options.print_product_states = vm.count("print-product-states") > 0;
    options.print_product_successors = vm.count("print-product-successors") > 0;
    if (options.print_product_successors) {
        // this option implies printing the product states as well
        options.print_product_states = true;
    }
    options.verbose_solver = vm.count("verbose-solver") > 0;

    // misc options
    if (vm.count("classpath")) {
        options.classpath = vm["classpath"].as<std::string>();
    }
    else {
        // look for Owl library in same directory as binary
        boost::filesystem::path full_path(boost::filesystem::initial_path<boost::filesystem::path>());
        full_path = boost::filesystem::system_complete(boost::filesystem::path(argv[0]));
        boost::filesystem::path owl_jar = full_path.parent_path() / "owl.jar";
        options.classpath = owl_jar.string();
    }
    options.timing = vm.count("timing") > 0;
    options.help = vm.count("help") > 0;

    return options;
}

}
