#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <vector>
#include <memory>
#include <thread>

#include "OptionParser.h"
#include "Definitions.h"
#include "util/Timer.h"
#include "ltl/LTLParser.h"
#include "ltl/TLSFParser.h"
#include "mealy/MealyMachine.h"
#include "pg/PGArena.h"
#include "pg/PGSolver.h"
#include "pg/PGSISolver.h"
#include "aig/AigerConstructor.h"

#include "owl.h"

void check_realizability(std::unique_ptr<pg::PGArena>& arena, std::unique_ptr<pg::PGSolver>& solver, strix::Options& options) {
    Timer timer(options.timing);

    timer.start("initializing OWL-JNI");
    std::stringstream classpath_sstream;
    classpath_sstream << "-Djava.class.path=" << options.classpath;
    const std::string classpath_option = classpath_sstream.str();
    owl::OwlJavaVM owlJavaVM(classpath_option.c_str(), options.validate_jni, options.max_heap_size_gb, options.aggressive_heap_optimizations);
    owl::OwlThread owl = owlJavaVM.attachCurrentThread();
    timer.stop();

    timer.start("parsing input file");
    std::unique_ptr<ltl::Parser> parser;
    if (options.ltl) {
        parser = std::unique_ptr<ltl::Parser>(new ltl::LTLParser(owl));
    }
    else {
        parser = std::unique_ptr<ltl::Parser>(new ltl::TLSFParser(owl));
    }
    if (options.formula.empty()) {
        if (options.input_file.empty()) {
            throw std::invalid_argument("No input file or formula was given.");
        }
        else {
            options.formula = parser->parse_file(options.input_file);
        }
    }
    owl::Formula formula = parser->parse_string(options.formula, options.inputs, options.outputs, options.print_formula);
    parser.reset();
    timer.stop();

    timer.start("constructing automaton tree");
    {
        owl::EmersonLeiAutomaton automaton = owl.createAutomaton(formula, options.simplify_formula, options.monolithic, options.splitting, options.onthefly);
        aut::AutomatonTreeStructure structure(std::move(automaton));
        if (options.print_tree) {
            structure.print(options.print_formula);
        }
        arena = std::unique_ptr<pg::PGArena>(new pg::PGArena(options.inputs.size(), options.outputs.size(), std::move(structure)));
    }
    timer.stop();

    if (options.onthefly) {
        timer.start("solving game in parallel with construction of arena");
    }
    else {
        timer.start("constructing arena");
        arena->constructArena(false, options.print_product_states, options.print_product_successors);
        timer.stop();
        timer.start("solving game");
    }
    solver = std::unique_ptr<pg::PGSolver>(new pg::PGSISolver(*arena, options.onthefly, options.threads, options.compact_colors, options.verbose_solver));
    if (options.onthefly) {
        std::thread solver_thread = std::thread(&pg::PGSolver::solve, &*solver);
        arena->constructArena(true, options.print_product_states, options.print_product_successors);
        solver_thread.join();
    }
    else {
        solver->solve();
    }
    timer.stop();
}

void construct_solution(std::unique_ptr<pg::PGArena>& arena, std::unique_ptr<pg::PGSolver>& solver, strix::Options& options, std::ostream& out) {
    Timer timer(options.timing);

    timer.start("constructing Mealy machine");
    mealy::MealyMachine m(std::move(options.inputs), std::move(options.outputs));
    bool with_labels = solver->constructMealyMachine(m, options.labels);
    timer.stop();

    // arena and solver no longer needed, clear memory
    solver.reset();
    arena.reset();

    if (options.labels && !with_labels) {
        options.labels = false;
        std::cerr << "Error: cannot construct product labels" << std::endl;
    }

    if (options.minimize) {
        timer.start("minimize Mealy machine");
        if (!m.minimizeMachine(options.labels)) {
            options.minimize = false;
        }
        timer.stop();
    }
    if (options.mealy) {
        if (options.dot) {
            m.print_dot(out, options.labels, options.minimize);
        }
        else {
            m.print_kiss(out, options.labels, options.minimize);
        }
    }
    else {
        timer.start("constructing AIGER circuit");
        std::shared_ptr<aig::AigerConstructor> aiger;
        if (options.auto_aiger) {
            aiger = aig::AigerConstructor::findMinimalAiger(m, options.labels, options.minimize, options.bdd);
        }
        else {
            aiger = std::shared_ptr<aig::AigerConstructor>(new aig::AigerConstructor(m, options.labels, options.minimize));
            aiger->constructAndCompress(options.bdd);
        }
        if (options.bdd) {
            aiger->printBDDs(out);
        }
        else if (options.dot) {
            aiger->print_dot(out);
        }
        else {
            aiger->print_aiger(out, options.binary);
        }
        timer.stop();
    }
}

void synthesis(strix::Options& options) {
    Timer timer(options.timing);
    std::unique_ptr<pg::PGArena> arena;
    std::unique_ptr<pg::PGSolver> solver;

    timer.start("checking realizability");
    check_realizability(arena, solver, options);
    timer.stop();

    Player winner = solver->getWinner();
    switch (winner) {
        case SYS_PLAYER:
            std::cout << "REALIZABLE" << std::endl;
            break;
        case ENV_PLAYER:
            std::cout << "UNREALIZABLE" << std::endl;
            break;
        case UNKNOWN:
            std::cout << "UNKNOWN" << std::endl;
            break;
    }

    if (options.print_arena_info) {
        arena->print_basic_info();
    }

    std::streambuf* buf;
    std::ofstream of;
    if (options.output_file.empty()) {
        buf = std::cout.rdbuf();
    }
    else {
        of.open(options.output_file, std::ios::out);
        if (!of) {
            throw std::invalid_argument("Could not open output file: " + options.output_file);
        }
        buf = of.rdbuf();
    }
    std::ostream out(buf);

    if (options.parity_game) {
        timer.start("writing parity game");
        arena->print(out, winner);
        timer.stop();
    }
    else if (winner == SYS_PLAYER && !options.realizability) {
        timer.start("constructing solution");
        construct_solution(arena, solver, options, out);
        timer.stop();
    }

    of.close();
}

int main(const int argc, const char* argv[]) {
    try {
        // Parse options
        strix::OptionParser option_parser(argc, argv);
        strix::Options options = option_parser.parse_options();

        if (options.help) {
            option_parser.print_help();
            return EXIT_SUCCESS;
        }
        else {
            synthesis(options);
        }

    }
    catch (const std::exception& e) {
        std::cerr << "Error: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const std::string& e) {
        std::cerr << "Error: " << e << std::endl;
        return EXIT_FAILURE;
    }
    catch (...) {
        std::cerr << "Error: Unknown error." << std::endl;
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
